from colour_of_molecule.classes import File
from colour_of_molecule.utils.energy_units import energy_units, get_current_energy_units
from colour_of_molecule.analysis.common_tools import wrap_from_both_sides
from colour_of_molecule.plotting.plot_spectra import *
import sys, os


def file_in(path):
    return File(path)


if len(sys.argv) > 1:
    script_name = os.path.basename(sys.argv[0])
    script_file_path = os.path.abspath(script_name)
    args = sys.argv[1:]

    if "--save" in args or "-s" in args:
        from shutil import copy

        path0 = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(path0, "templates")

        all_templates = [f for f in os.listdir(path) if ".py" in f and f != "__init__.py"]

        def check_template_format(path):
            with open(path, 'r') as script:
                for line in script:
                    if "#" in line and "category:" in line and len(line.split(":")[-1].replace(" ","")) > 1:
                        return True
                    break

        if check_template_format(script_file_path) is not True:
            raise Exception("ERROR:\tScript couldn't be archived. The first line of the script has to contain \"### category: type_your_category_name_here\"")

        if ".py" not in script_name:
            raise Exception("ERROR:\tScript couldn't be archived. It doesn't seem to have the \".py\" extension.")

        while script_name in all_templates:

            print("WARNING:\tScript called"+wrap_from_both_sides(script_name, " \"")+"is already present in templates.")
            overwrite = input("\nDo you want to overwrite it? [Y/n] \n")

            if overwrite == "Y":
                break

            elif overwrite == "n":
                rename = input("\nDo you want to save it with different name? [Y/n] \n")

                if rename == "Y":
                    new_script_name = input("\nEnter new script name: \n") or script_name.replace(".py", "_new.py")

                    if ".py" in new_script_name:
                        script_name = new_script_name
                    else:
                        print("ERROR:\tInvalid file name. It doesn't seem to have the \".py\" extension. Reinitializing ...\n")
                        continue

                else:
                    print("INFO:\tScript archiving canceled.")
                    sys.exit(0)

            else:
                print("INFO:\tArgument hasn't been recognised. Script archiving canceled.")
                sys.exit(0)


        print("INFO:\tFile" + wrap_from_both_sides(script_name, " \"") + "will be copied"+
              "\n\t> from" + wrap_from_both_sides(script_file_path, " \"") +
              "\n\t> to" + wrap_from_both_sides(os.path.join(path, script_name), " \"")
              )
        confirmation = input("\nPress Enter to proceed. \n")

        if confirmation == "":
            target_path = os.path.join(path, script_name)
            copy(script_file_path, target_path)
            print("INFO:\tScript successfully archived in"+wrap_from_both_sides(target_path, " \""))
            sys.exit(0)

        else:
            print("INFO:\tScript archiving canceled.")
            sys.exit(0)






import re
from colour_of_molecule.classes import AbsLine, File, Energy, OpList
from colour_of_molecule.analysis.common_tools import homo_lumo, remove_dashes, wrap_from_both_sides
from colour_of_molecule.utils.energy_units import get_current_energy_units


class Configuration:
    def __init__(self, indices, config_a, config_b, amplitude):
        self.indices = indices
        self.alpha = OpList(config_a)
        self.beta = OpList(config_b)
        self.amplitude = amplitude
        self.length = self.get_length()

    def get_length(self):
        la = len(self.alpha)
        lb = len(self.beta)
        if la != lb:
            raise Exception("ERROR:\tConfigurations for alpha an beta electrons differ in length for a given state.")
        return la


class State:
    def __init__(self, number, multiplicity, energy):
        self.number = number
        self.multiplicity = multiplicity
        self.energy = Energy(energy, "eV")
        #self.conf = list()
        #
        self.configurations = list()

        self.MO_indices = list()

    def add_config_line(self, MO_indices, config_a, config_b, amplitude):
        self.configurations.append(Configuration(MO_indices, config_a, config_b, amplitude))
        self.MO_indices = MO_indices


def create_abs_lines(state1, state2, fosc, wavelength, homo):
    initial = state1.configurations[0]
    finals = state2.configurations

    def determine_change_in_occ(initial, final):
        change_in_occ = list(map(lambda i, j: len(i.replace("-", "")) - len(j.replace("-", "")), final, initial))
        return change_in_occ

    def determine_MOs_of_transition(change_in_occ, MO_ids, homo_id):
        trans_from = [homo_lumo(homo_id, y) for x, y in zip(change_in_occ, MO_ids) if x < 0]
        trans_to = [homo_lumo(homo_id, y) for x, y in zip(change_in_occ, MO_ids) if x > 0]
        return trans_from, trans_to

    def ids_to_MO_names(initial, final, MO_ids, homo_id):
        dif_a = determine_change_in_occ(initial.alpha, final.alpha)
        dif_b = determine_change_in_occ(initial.beta, final.beta)
        output = list()

        for x in dif_a, dif_b:
            if sum([abs(y) for y in x]) != 0:
                output.append(list(determine_MOs_of_transition(x, MO_ids, homo_id)))
                output[0].append(final.amplitude)
        return output

    transitions = list()
    for cf in finals:
        data = ids_to_MO_names(initial, cf, state1.MO_indices, homo)
        if len(data) != 0:
            transitions.append(data[0])

    output = AbsLine(wavelength, fosc, transitions)
    return output


class File_mndo(File):
    class AbsLines_getter:
        def __get__(self, instance, owner):
            data = Mndo_file_to_abslines(instance)
            return data

    abs_lines = AbsLines_getter()


def Mndo_file_to_abslines(log_file):
    shift = log_file._shift

    with open(log_file.path, "r") as file:
        states = dict()
        config = list()
        output = list()

        reg_state = re.compile("^\s?State\s*\d+,")
        reg_line = re.compile("^\s?-+")
        reg_num = re.compile("[\d.]+")
        reg_ab = re.compile("ab|[ab-]{1}")

        def filter_ab(lst, ab):
            out = [x.replace("-", "").replace(ab, "") for x in lst]
            return out

        read = False
        transitions = False

        state_no = 0
        empty_line = 0
        empty_line2 = 0
        init_state = int()

        for line in file:
            ### first part - first tables with states descriptions
            if reg_state.search(line) is not None:
                nums = reg_num.findall(line)
                state_no = int(nums[0])
                states.update({state_no: State(state_no, nums[2], nums[4])})
                empty_line = 0

            if read is True:
                if reg_num.search(line) is not None:
                    numerals = reg_num.findall(line)
                    amplitude = float(numerals[1])
                    indices = numerals[3:]

                elif reg_ab.search(line) is not None:
                    config = reg_ab.findall(line)
                    config_a = filter_ab(config, "a")
                    config_b = filter_ab(config, "b")
                    states.get(state_no).add_config_line(indices, config_a, config_b, amplitude)

                    config = []
                    config_a = []
                    config_b = []
                    indices = []
                    amplitude = 0
                    empty_line = 0

                else:
                    empty_line += 1

            if empty_line == 2:
                read = False

            if state_no != 0 and reg_line.search(line) is not None and empty_line != 2:
                read = True


            ### second part - second table
            if empty_line2 >2:
                transitions = False

            if "Properties of transitions" in line:
                transitions = True
                init_state = int(reg_num.findall(line)[0])

            elif transitions is True:
                if "Symmetry" not in line and reg_num.search(line) is not None:
                    numerics = [float(x) for x in reg_num.findall(line)]
                    homo_id = max([i for i, j in zip(states.get(1).MO_indices, states.get(1).configurations[0].alpha) if len(j) > 0])
                    energy = Energy(numerics[4], "nm") + shift
                    wavelength = energy.in_units("nm")
                    abline = create_abs_lines(states.get(init_state), states.get(int(numerics[0])), numerics[5], wavelength, homo_id)
                    output.append(abline)

                elif "Symmetry" not in line:
                    empty_line2 += 1

    if shift.value != 0:
        print("Shift applied to wavelengths:\n  ", shift, " nm")

    return output




### category: Colours_and_plotting

import sys
import colour_of_molecule as com

# Path to the file can be entered here:
input_file = "C:/"
save_path = ""


# Accepting command line arguments:
if len(sys.argv) > 1:
    path = sys.argv[1]
else:
    path = input_file


# Loading the file:
file = com.file_in(path)


# Settings:
with com.energy_units("nm"):
    file.plot_range = (100, 1000)

with com.energy_units("cm-1"):
    file.fwhm = 1000


# Font settings
from colour_of_molecule.classes.classes import FontSettings
fonts = FontSettings(newfonts={'all': 'Arial'}, newsizes={'title': 11, 'legend': 8}, use_all=True)


# Generate output filenames
if save_path == "":
    from colour_of_molecule.analysis.common_tools import get_output_path, file_saver
    save_path = get_output_path(path, name="{0}_IMG_colmol.png".format(file.filename.split(".")[0]))


# Generating output:
with com.energy_units("cm-1"):
    com.plot_single_spectrum(file, save=file_saver(save_path), fonts=fonts)
    com.plot_abs_lines(file, save=file_saver(save_path), fonts=fonts)






import os
from tracemalloc import start
import numpy as np
from ..utils.energy_units import get_current_energy_units, convert_nm_to_rcm

class AbsLine:
    def __init__(self, energy, strength, *transitions):
        self.energy = Energy(energy, "nm")
        self.oscillator_strength = strength
        self.transitions = list(transitions)


class MolarAbsSpectrum:
    def __get__(self, instance, owner):
        from colour_of_molecule.analysis.spectrum import abslines_to_molar_abs

        data = abslines_to_molar_abs(instance)
        return data

    
class Transmittance:
    def __get__(self, instance, owner):
        from colour_of_molecule.analysis.spectrum import molar_abs_to_transmittance

        data = molar_abs_to_transmittance(instance.molar_abs_spectrum, OD=instance.optical_density)
        return data


class ColourRGB:
    def __get__(self, instance, owner):
        from colour_of_molecule.analysis.spectrum import find_colour
        rgb = find_colour(instance.complementary_abs_spectrum)
        return rgb


class Energy:
    __default_unit = "kcal/mol"

    supported_units = {  # 1 unit = ... kcal/mol
        "kcal/mol": 1,
        "Hartree": 630,
        "au": 630,
        "eV": 23.060541945329334,
        "kJ/mol": 0.2390057361376673,
        "Ry": 315,
        "cm-1": 0.0028591441663694465,
        "nm": convert_nm_to_rcm,
        "Hz": 9.53707836896099e-14,
        "GHz": 9.53707836896099e-8
    }

    def __init__(self, value, units=__default_unit):
        self.units = self.__default_unit

        if units is self.__default_unit:
            self.value = value
        elif units in self.supported_units:
            if callable(self.supported_units.get(units)) and units == "nm":
                func = self.supported_units.get(units)
                coeff = self.supported_units.get("cm-1") / self.supported_units.get(self.__default_unit)
                val = func(value) * coeff
                if val.size == 1:
                    self.value = val.item()
                else:
                    self.value = np.array(val)
            else:
                coeff = self.supported_units.get(units) / self.supported_units.get(self.__default_unit)
                self.value = value * coeff
        else:
            raise Exception("ERROR:\tUnsupported unit was encountered. Only these units are currently supported: " +
                            "".join([str(i) + "  " for i in self.supported_units.keys()]))

    def __sub__(self, other):
        result = self.value - other.value
        return Energy(result)

    def __add__(self, other):
        result = self.value + other.value
        return Energy(result)
    
    def __repr__(self):
        return f"{self.in_units(get_current_energy_units())} {get_current_energy_units()}"

    def in_units(self, new_units):
        if new_units == "nm":
            out = convert_nm_to_rcm(self.value / self.supported_units.get("cm-1") * self.supported_units.get(self.__default_unit))
            if out.size == 1:
                out = out.item()
            return np.round(out, 4)
        
        elif new_units in self.supported_units:
            out = self.value / self.supported_units.get(new_units) * self.supported_units.get(self.__default_unit)
            return out
        
        else:
            raise Exception("ERROR:\tUnsupported unit was encountered. Only these units are currently supported: " +
                            "".join([str(i) + "  " for i in self.supported_units.keys()]))
        
    def in_current_units(self):
        return self.in_units(get_current_energy_units())


class EnergyAxis:
    def __get__(self, instance, owner):
        start = int(instance.start.in_units("nm"))
        end = int(instance.end.in_units("nm"))
        npoints = end - start + 1
        return Energy(np.linspace(start, end, npoints), units="nm")


class EnergyRange(Energy):
    def __init__(self, start, end, units=Energy._Energy__default_unit):
        if isinstance(start, Energy) and isinstance(end, Energy):
            start = start.value
            end = end.value
            units = Energy.__default_unit

        super().__init__(np.array([start, end]), units)
        self.start = Energy(self.value[0])
        self.end = Energy(self.value[1])

    def __repr__(self):
        start = self.start.in_current_units()
        end = self.end.in_current_units()
        sorted_range = sorted((start, end))
        return f"({sorted_range[0]}, {sorted_range[1]}) {get_current_energy_units()}"

    energy_axis = EnergyAxis()


class OpList(list):
    def __add__(self, other):
        return OpList([x + y for (x, y) in zip(self, other)])

    def __sub__(self, other):
        return OpList([x - y for (x, y) in zip(self, other)])

    def __neg__(self):
        return OpList([-x for x in self])

    def get_abs_length(self):
        return int(sum([abs(x) for x in self]))

    def as_lengths(self):
        return OpList([len(x) for x in self])


class FontSettings():
    def __init__(self, newfonts=list(), newsizes=list(), use_all=False):
        self.fontdict = {'title': 'Calibri',
                    'axis': 'Calibri',
                    'axis_tick_labels': 'Calibri',
                    'legend':'Calibri',
                    'all':'Calibri'}
        self.sizedict = {'title': 14, 'axis': 12, 'axis_tick_labels': 12, 'legend': 12, 'all': 12}
        self.fontdict.update(newfonts)
        self.sizedict.update(newsizes)

        if use_all is True:
            self.fontdict = {key:self.fontdict['all'] for key in self.fontdict.keys()}

        import matplotlib.font_manager as font_manager
        self.fonts = {key:font_manager.FontProperties(family=self.fontdict[key], weight='normal', style='normal',
                                                      size=self.sizedict[key])
                 for key in self.fontdict.keys()}


class File:
    supported_formats = {"gaussian": (0, "Entering Link 1"),
                         "orca": (3, "* O   R   C   A *"),
                         "mndo": (0, "PROGRAM MNDO"),
                         "molpro": (0, "***  PROGRAM SYSTEM MOLPRO  ***")
                         }

    def __init__(self, path):
        def check_line(dict, line):
            ans = [k for k in list((i, j[0]) if j[1] in line else False for i, j in dict.items()) if k]
            if ans:
                return ans[0]

        def sanity_check_type_module_class(self):
            if not hasattr(self, "type"):
                raise Exception("ERROR:\tAttribute \"type\" not found.")
            else:
                from importlib.util import find_spec

                mod_path = "colour_of_molecule.input." + self.type
                class_name = "File_" + self.type

                check_module = find_spec(mod_path)
                if check_module is None:
                    raise Exception("ERROR:\tModule \"" + mod_path + "\" not found. Is this file type implemented yet?")
                else:
                    class_check = hasattr(import_module(mod_path), class_name)
                    if class_check is not True:
                        raise Exception("ERROR:\tClass \"" + class_name + "\" not found in module \"" + mod_path + "\"")
            #print("INFO:\tAll sanity checks passed successfully.")
            pass

        self.path = path
        self.filename = os.path.basename(path)
        self.ranges_of_comps = dict()
        self.number_of_comps = 0
        self.more_than_one_comp = False

        self._fwhm = Energy(1000.0, "cm-1")
        self._standard_deviation = Energy(self._fwhm.value / 2.3548)
        self._optical_density = 0.15
        self._plot_range = EnergyRange(200, 800, "nm")
        self.transition_minimal_amplitude = 0.5
        self.normalize_absorption_spectrum = True
        self._shift = Energy(0.0)

        self.plot_title = ""
        self.legend_title = ""

        with open(path, "r") as file:

            self.name = os.path.basename(file.name).replace("_", "-")

            for index, line in enumerate(file):
                out = check_line(self.supported_formats, line)
                if out:
                    self.type = out[0]
                    self.ranges_of_comps.update({self.number_of_comps: (index - out[1], None)})
                    if self.number_of_comps > 0:
                        self.more_than_one_comp = True
                        self.ranges_of_comps.update({self.number_of_comps -1 : (self.ranges_of_comps.get(self.number_of_comps-1)[0], index - out[1] - 1)})
                    self.number_of_comps += 1

            if self.type is None:
                raise Exception("ERROR:\tInput file type is not implemented.\n\tOnly these types are currently supported:  " +
                                "".join([str(i).upper()+"  " for i in self.supported_formats.keys()]))
            print("INFO:\tNumber of recognised computations in "+self.type.capitalize()+" file \""+self.name+"\" is:   "+str(self.number_of_comps))

            ### Switch for multiple files should be here!!!  ###

            from importlib import import_module

            sanity_check_type_module_class(self)

            print("INFO:\tChanging class to \"File_"+self.type+"\"")
            clss = getattr(import_module("colour_of_molecule.input."+self.type), "File_"+self.type)
            self.__class__ = clss

    # optical density
    @property
    def optical_density(self):
        return self._optical_density

    @optical_density.setter
    def optical_density(self, value):
        self._optical_density = value
    
    # FWHM & stdev
    @property
    def fwhm(self):
        if get_current_energy_units() == "nm":
            raise Exception("ERROR:\tFWHM cannot be retrieved in 'nm' without specifying reference energy. Use get_corrected_fwhm(reference_energy) method instead.")
        else:
            return self._fwhm.in_current_units()
    
    @fwhm.setter
    def fwhm(self, value):
        if get_current_energy_units() == "nm":
            raise Exception("ERROR:\tFWHM cannot be set in 'nm' units due to non-linear relationship between energy and wavelength.")
        else:
            self._fwhm = Energy(value, get_current_energy_units())

    def get_corrected_fwhm(self, reference_energy):
        if get_current_energy_units() == "nm":
            reference = Energy(reference_energy, get_current_energy_units())
            shifted = Energy(reference.value + self._fwhm.value)
            return abs(shifted.in_current_units() - reference.in_current_units())
        else:
            return self._fwhm.in_current_units()

    @property
    def standard_deviation(self):
        return self.fwhm / 2.3548
    
    @standard_deviation.setter
    def standard_deviation(self, value):
        self.fwhm = value * 2.3548

    def get_corrected_standard_deviation(self, reference_energy):
        corrected_fwhm = self.get_corrected_fwhm(reference_energy)
        return corrected_fwhm / 2.3548

    # range for plotting
    @property
    def plot_range(self):
        start = self._plot_range.start.in_current_units()
        end = self._plot_range.end.in_current_units()
        return sorted((start, end))
    
    @plot_range.setter
    def plot_range(self, value):
        if isinstance(value, tuple) and len(value) == 2:
            self._plot_range = EnergyRange(*value, get_current_energy_units())
        elif isinstance(value, EnergyRange):
            self._plot_range = value
        else:
            raise Exception("ERROR:\tPlot range must be set as a tuple of two values or as an EnergyRange instance.")

    # shift to be applied on the energy axis
    @property
    def shift(self):
        return self._shift.in_current_units()
    
    @shift.setter
    def shift(self, value):
        if get_current_energy_units() == "nm":
            raise Exception("ERROR:\tShift cannot be set in 'nm' units due to non-linear relationship between energy and wavelength.")
        else:
            self._shift = Energy(value, get_current_energy_units())

    def get_corrected_shift(self, reference_energy):
        if get_current_energy_units() == "nm":
            reference = Energy(reference_energy, get_current_energy_units())
            shifted = Energy(reference.value + self._shift.value)
            return abs(shifted.in_current_units() - reference.in_current_units())
        else:
            return self._shift.in_current_units()


    molar_abs_spectrum = MolarAbsSpectrum()
    transmittance = Transmittance()
    colour_rgb = ColourRGB()








def homo_lumo(HOMO_no, orbital_no):
    if int(orbital_no) <= int(HOMO_no):
        output = "HOMO"
        d = int(orbital_no) - int(HOMO_no)
    else:
        output = "LUMO"
        d = int(orbital_no) - int(HOMO_no) - 1
    if d > 0:
        output += "+"+str(d)
    elif d < 0:
        output += str(d)
    return output


def remove_dashes(lst):
    return list(map(lambda x: x.replace("-",""), lst))


def determine_number_of_calculations(file):
    type = file.type
    switch = {
        "gaussian": "",
        "orca": ""
    }
    pass


def normalize_list(lst):
    norm = max(lst)
    return [x/norm for x in lst]


def wrap_from_both_sides(middle, symbol="*"):
    return str(symbol) + str(middle) + str(symbol)[::-1]


def file_saver(path):
    import os
    number = 0

    while os.path.isfile(path) is True:
        pth, ext = path.rsplit('.', 1)

        if number == 0:
            path = "_{:d}.".format(number).join([pth, ext])
        else:
            path = "_{:d}.".format(number).join([pth[:-(len(str(number - 1)) + 1)], ext])
        number += 1

    return path


def get_output_path(input_path, name="IMG_col-of-mol.png"):
    import os
    dir = os.path.dirname(input_path)
    output = os.path.join(dir, name)
    return output



# def check_save():
#     import sys, os
#     from shutil import copy
#
#     if sys.argv[1] == 'save':
#         com_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#         templates_dir = os.path.join(com_dir, 'templates')
#         script_name = os.path.basename(sys.argv[0])
#         script_path = os.path.join(os.getcwd(), script_name)
#         target_path = os.path.join(templates_dir, script_name)
#
#         print(templates_dir)
#         print(script_path)
#
#         rename = input("\nDo you want to archive the script \"{}\" in package templates? [Y/n] \n".format(script_name))
#         if rename == "Y":
#             copy_path = file_saver(target_path)
#         else:
#             print("\nINFO:\tScript archiving canceled.")
#             sys.exit(0)
#
#         print("\nINFO:\tFile \"{}\" will be copied \n\t> from \"{}\" \n\t> to \"{}\"".format(script_name, script_path, copy_path))
#
#         confirmation = input("\nPress Enter to proceed. \n")
#
#         if confirmation == "":
#             copy(script_path, copy_path)
#             print("INFO:\tScript successfully copied to \"{}\"\n".format(copy_path))
#             sys.exit(0)
#
#         else:
#             print("INFO:\tScript archiving canceled.")
#             sys.exit(0)





import re
from colour_of_molecule.classes import AbsLine, File, Energy
from colour_of_molecule.analysis.common_tools import homo_lumo
from colour_of_molecule.utils.energy_units import get_current_energy_units


class File_gaussian(File):
    class AbsLines_getter:
        def __get__(self, instance, owner):
            data = Gauss_log_to_abslines(instance)
            return data

    abs_lines = AbsLines_getter()


def Gauss_log_to_abslines(log_file):
    file = open(log_file.path, "r")
    shift = log_file._shift

    list_file = list()
    fs = list()
    wav = list()

    parameters = list()
    # electrons = list()
    orbitals = list()
    all_orbitals = list()

    reg_exc = re.compile("Excited State")
    reg_wav = re.compile("\s[\d.]*\snm")
    reg_f = re.compile("f\=[\d.]*\s")
    reg_par = re.compile("^(\s?#)p\s")
    reg_line = re.compile("^(\s?-)-*")
    reg_charge = re.compile("^\s?Charge.*\sMultiplicity")
    reg_MOs = re.compile("\d+\s[(alpha)(beta)]+\selectrons")
    reg_num = re.compile("[\d.]+")

    ki = False
    ori = False
    HOMO = 0

    multiplicity = {0: "singlet",
                    1: "doublet",
                    2: "triplet",
                    3: "quartet",
                    4: "quintuplet"}

    for x in file:
        if reg_MOs.search(x) is not None and HOMO == 0:
            electrons = reg_num.findall(x)
            diff = abs(int(electrons[0]) - int(electrons[1]))

            if diff in multiplicity:
                pass
                #print("INFO:  Spin multiplicity is " + str(diff + 1) + ". It's a " + multiplicity[diff] + ".")
            else:
                pass
            HOMO = max(electrons)

        if ori is True:
            if len(reg_num.findall(x)) == 3:
                numerals = reg_num.findall(x)
                orb_nos = numerals[:2]
                amplitude = float(numerals[2])
                ltt = list(map(lambda z: homo_lumo(HOMO, z), orb_nos))
                ltt.append(amplitude)
                orbitals.append(ltt)
            else:
                all_orbitals.append(orbitals)
                orbitals = []
                ori = False

        if reg_exc.search(x) is not None and ori is False:
            list_file.append(x)
            ori = True

        if ki is True:
            if reg_line.search(x) is None:
                parameters.append(x)
            else:
                ki = False

        if reg_par.search(x) is not None:
            parameters.append(x)
            ki = True

        if reg_charge.search(x) is not None:
            ch_mult = x[1:]

        if reg_MOs.search(x) is not None:
            reg_num.findall(x)
    file.close()

    header = ""
    for q in parameters:
        header = header + q[1:len(q) - 1]

    for y in list_file:
        loc_wav = reg_wav.search(y)
        loc_f = reg_f.search(y)
        wavelength = Energy(float(y[loc_wav.start() + 1: loc_wav.end() - 3]), 'nm')
        # energy shift needs to be applied in linear energy units, not wavelength units
        shifted_wavelength = wavelength + shift
        wav.append(shifted_wavelength.in_units("nm"))
        fs.append(float(y[loc_f.start() + 2: loc_f.end() - 1]))

    if wav is []:
        raise Exception("Error in file import. Check the encoding of .txt file and eventually change it to ANSI.")

    if shift.value != 0:
        print(f"Shift of {shift.in_current_units()} {get_current_energy_units()} was applied to the energy axis:\n")

    output = list(map(lambda i, j, k: AbsLine(i, j, k), wav, fs, all_orbitals))

    return output



import re
from colour_of_molecule.classes import AbsLine, File, Energy
from colour_of_molecule.analysis.common_tools import homo_lumo, wrap_from_both_sides


class DipoleTransMoment:
    def __init__(self):
        self.dmx = 0
        self.dmy = 0
        self.dmz = 0
        self.energy_difference = float()
        self.oscillator_strength = float()

    def get_oscillator_strength(self, dE):
        if isinstance(dE, float):
            self.energy_difference = dE
        else:
            raise Exception("ERROR:\tArgument has to be either a float (in atomic units) or a tuple of two \"Energy\" class instances.")

        self.oscillator_strength = 2/3 * self.energy_difference * (self.dmx**2 + self.dmy**2 + self.dmz**2)

        return self.oscillator_strength


class SingleState:
    def __init__(self, number, energy):
        self.number = float(number)
        self.energy = Energy(energy, "au")


class StateTransition:
    def __init__(self, num_final, energy, num_initial=1.1):
        self.num_final = float(num_final)
        self.num_initial = float(num_initial)
        self.symmetry = int(str(num_final).split(".")[1])
        self.energy = Energy(energy, "au")
        self.wavelength = self.energy.in_units("nm")
        self.osc_strength = float()
        self.MO_indices = list()

        self.dipole_trans_moment = DipoleTransMoment()

    def add_MO_transition(self, mo1, mo2, coeff):
        self.MO_indices.append([float(mo1), float(mo2), float(coeff)])

    def add_osc_strength(self, fosc):
        self.osc_strength = float(fosc)


class MainReader:
    reg_num = re.compile("-?\d+[.\d]*")
    reg_alphabet = re.compile("[a-zA-Z]+")
    reg_res_for_states = re.compile("^\s*Results for state\s*\d+\.\d+")
    reg_oscillator = re.compile("^\s*Oscillator strength")
    reg_orbitals = re.compile("^\sELECTRON ORBITALS")


class EomReader(MainReader):
    reg_eom_final = re.compile("^\s*Final Results for EOM")

    def __call__(self, log_file):
        shift = log_file._shift

        with open(log_file.path, "r") as file:

            state_transitions = dict()
            index = float()
            id_homo = float()
            #id_lumo = float()
            listlong = list()
            longest = int()
            c_empty_line = 0
            c_results_for_states = False
            eom_final = False
            c_oscill = False
            c_orbitals = False

            for line in file:
                # first part - states and transitions
                if self.reg_res_for_states.search(line) is not None:
                    index, eng = [float(x) for x in self.reg_num.findall(line)[:2]]
                    state_transitions.update({
                        index: StateTransition(index, eng)
                    })
                    c_results_for_states = True
                    c_empty_line = 0

                elif c_results_for_states is True and index > 0:
                    if self.reg_num.search(line) is not None and "->" in line:
                        mos = self.reg_num.findall(line)
                        state_transitions.get(index).add_MO_transition(mos[1], mos[2], mos[0])
                        c_empty_line = 0

                    else:
                        c_empty_line += 1

                if c_empty_line > 3:
                    c_results_for_states = False

                # second part - transition dipole moments
                if self.reg_eom_final.search(line) is not None:
                    eom_final = True

                if eom_final is True:
                    if self.reg_oscillator.search(line) is not None:
                        c_oscill = True

                    elif c_oscill is True and self.reg_num.search(line) is not None:
                        nums = [float(x) for x in self.reg_num.findall(line)]
                        state_transitions.get(nums[0]).add_osc_strength(nums[2])
                        c_oscill = False

                # third part - MO indexes, names
                if self.reg_orbitals.search(line) is not None:
                    c_orbitals = True

                if c_orbitals is True and self.reg_alphabet.search(line) is None:
                    length = len(self.reg_num.findall(line))
                    if length > longest and length > 0:
                        longest = length
                        listlong = []
                        listlong.append(self.reg_num.findall(line))
                    elif length == longest:
                        listlong.append(self.reg_num.findall(line))

                if " HOMO " in line:
                    c_orbitals = False
                    id_homo = float(self.reg_num.findall(line)[0])
                #
                # if " LUMO " in line:
                #     c_orbitals = False
                #     id_lumo = float(self.reg_num.findall(line)[0])

        def format_listlong(listlong):
                if not isinstance(listlong, list):
                    raise Exception("ERROR:\tArgument has to be a list.")

                sortlist = list()

                for x in listlong:
                    sortlist.append([float(x[2]), float(x[0]), int(x[1])])

                out = list(enumerate(sorted(sortlist)))

                return {y[1]: x for (x, y) in out}

        def homo_lumo_for_MO(dct, homo, mos):
            if not isinstance(dct, dict):
                raise Exception("ERROR:\tFirst argument has to be a dictionary of floats in both key and value.")
            if not isinstance(id_homo, float):
                raise Exception("ERROR:\tSecond argument has to be a float (orbital.symmetry).")
            if not isinstance(mos, list) and len(mos) != 2:
                raise Exception("ERROR:\tThird argument has to be a list of two floats (initial and final orbital.symmetry).")

            def get_and_check(dct, homo_id, id):
                homo = dct.get(homo_id)

                if dct.get(id) is None:
                    return wrap_from_both_sides(id, symbol="**")
                    #return "*"+str(id)+"*"
                else:
                    return homo_lumo(homo, dct.get(id))

            out = [
                get_and_check(dct, homo, mos[0]),
                get_and_check(dct, homo, mos[1]),
                abs(mos[2])
                   ]
            return out

        dictlong = format_listlong(listlong)
        output = list()

        for transition in state_transitions.values():
            energy = Energy(transition.wavelength, "nm") + shift
            output.append(AbsLine(energy.in_units("nm"),
                                  transition.osc_strength,
                                  [homo_lumo_for_MO(dictlong, id_homo, x) for x in transition.MO_indices]
                                  )
                          )
        return output


class MrciReader(MainReader):
    reg_mrci_energy = re.compile("^\s*!MRCI STATE \d+\.\d+ Energy")
    reg_mrci_trans = re.compile("^\s*!MRCI trans")
    reg_mrci_DMX = re.compile("\|DMX\|")
    reg_mrci_DMY = re.compile("\|DMY\|")
    reg_mrci_DMZ = re.compile("\|DMZ\|")

    def __call__(self, log_file):
        shift = log_file._shift
        states = dict()
        state_transitions = dict()

        with open(log_file.path, "r") as file:
            for line in file:
                if self.reg_mrci_energy.search(line) is not None:
                    nums = [float(x) for x in self.reg_num.findall(line)]
                    states.update({nums[0]: SingleState(nums[0], nums[1])})

                if self.reg_mrci_trans.search(line) is not None:
                    nums_in_string = [float(x) for x in self.reg_num.findall(line)[:2]]

                    if len([x for x in list(states.keys()) if x in nums_in_string]) == 2:
                        state1, state2 = sorted([float(x) for x in nums_in_string])[:2]

                        trans_id = (state1, state2)
                        eng1 = states.get(state1).energy
                        eng2 = states.get(state2).energy
                        delta_eng = eng2 - eng1

                        state_transitions.update({
                            trans_id: StateTransition(state2, delta_eng.in_units("au"), state1)
                        })

                        if self.reg_mrci_DMX.search(line) is not None:
                            dmx = self.reg_num.findall(line)[2]
                            state_transitions.get(trans_id).dipole_trans_moment.dmx = float(dmx)

                        if self.reg_mrci_DMY.search(line) is not None:
                            dmy = self.reg_num.findall(line)[2]
                            state_transitions.get(trans_id).dipole_trans_moment.dmy = float(dmy)

                        if self.reg_mrci_DMZ.search(line) is not None:
                            dmz = self.reg_num.findall(line)[2]
                            state_transitions.get(trans_id).dipole_trans_moment.dmz = float(dmz)

        output = list()

        for transition in state_transitions.values():
            energy = transition.energy + shift
            output.append(AbsLine(energy.in_units("nm"),
                                  transition.dipole_trans_moment.get_oscillator_strength(energy.in_units("au")),
                                  [[wrap_from_both_sides(transition.num_initial), wrap_from_both_sides(transition.num_final)]]
                                  )
                          )
        return output


class Rs3Reader(MainReader):
    def __call__(self, log_file_path):
        print("ERROR:   Rs3 type is not implemented yet.")
        pass
    pass


class File_molpro(File):
    class AbsLines_getter:
        def __get__(self, instance, owner):
            typ, reader = instance.molpro_subtype
            the_reader = reader()
            data = the_reader(instance)
            return data

    class MolproSubtype:
        def __get__(self, instance, owner):
            def check_tuple_re_func(tup):
                from re import Pattern

                if not isinstance(tup, tuple) and len(tup) != 2:
                    raise Exception("ERROR:\tDictionary has to have format \"a: (re.Pattern, function)\".")

                reg, func = tup

                if not isinstance(reg, Pattern):
                    raise Exception("ERROR:\tDictionary has to have format \"a: (re.Pattern, function)\".")

                if not callable(func):
                    raise Exception("ERROR:\tDictionary has to have format \"a: (re.Pattern, function)\".")

            def determine_subtype(lin, supported):
                if not isinstance(supported, dict):
                    raise Exception("ERROR:\tWrong second argument type. Has to be \"dict\".")

                out = tuple()
                for typ, tools in supported.items():
                    check_tuple_re_func(tools)

                    reg_type, func = tools

                    if reg_type.search(lin) is not None:
                        out = (typ, func)
                if out == ():
                    pass
                else:
                    return out

            with open(instance.path, 'r') as file:
                for line in file:
                    found = determine_subtype(line, instance.supported_subtypes)
                    if found is not None:
                        typ, func = found
                        break
            return typ, func

    abs_lines = AbsLines_getter()
    molpro_subtype = MolproSubtype()

    supported_subtypes = {
        "mrci": (re.compile("^\s?\{?ci"), MrciReader),
        "eom": (re.compile("^\s?\{?eom"), EomReader),
        "rs3": (re.compile("^\s?\{?rs3"), Rs3Reader)
    }






from os import listdir
import re, os, sys
from shutil import copy
from colour_of_molecule.analysis.common_tools import wrap_from_both_sides


def update_dict(dictionary, key, value):
    if dictionary.get(key):
        dictionary[key].append(value)
    else:
        dictionary[key] = [value]


def interactive_cmd(opt1, *args):
    from InquirerPy import prompt, inquirer

    def add_points(lst):
        for j in lst:
            key = list(j.keys())[0]
            val = "> " + list(j.values())[0]
            j.update({key: val})
        return lst

    opt1 = add_points(opt1)
    back = True

    while back is True:
        exit_option = [{"name": "-- exit", "value": "exit"}]
        question1 = [
            {
                "type": "list",
                "message": "What category of templates are you interested in?  (use arrows to navigate)",
                "choices": opt1 + exit_option,
            }, ]
        result1 = prompt(questions=question1)

        if result1[0] == "exit":
            raise IOError

        options2 = args[result1[0]]

        back_option = [{"name": "-- back", "value": True}]
        question2 = [
            {
                "type": "list",
                "message": "Choose template you wish to create.  (use arrows to navigate)",
                "choices": options2 + back_option
            }, ]

        result2 = prompt(questions=question2)

        if result2[0] is not True:
            back = False

    return result1[0], result2[0]


def static_cmd(opt1, *args):
    def print_static(level_1, *args):
        print("Select a template you wish to import:\n")
        counter = 0
        output = dict()
        for i in level_1:
            label = str(i.get("name"))
            index = i.get("value")
            print(label)
            for k in args[index]:
                print("    " + str(counter) + "  >  " + k.get("name"))
                output.update({counter: k})
                counter += 1

        print("\nThen run a function \"colour_of_molecule.templates.create(#)\" where # is the index of selected file to copy it into current directory.")
        return output

    global indexed_templates
    indexed_templates = print_static(opt1, *args)

    global create
    def create(index, path, list=indexed_templates):
        answ = list.get(index).get("name")
        template_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), answ)
        print("path: ", path)
        if "." in path and ".py" in path:
            print("INFO:\tCopying template" + wrap_from_both_sides(answ, " \"") +
                  "\n\t> from" + wrap_from_both_sides(template_path, " \"") +
                  "\n\t> to" + wrap_from_both_sides(path, " \"")
                  )
            copy(template_path, path)
        else:
            target_path = os.path.join(os.path.dirname(path), answ)
            print("INFO:\tCopying template"+wrap_from_both_sides(answ, " \"")+
                  "\n\t> from"+wrap_from_both_sides(template_path, " \"")+
                  "\n\t> to"+wrap_from_both_sides(target_path, " \"")
                  )
            copy(template_path, target_path)

        return answ


### always running:

path = os.path.dirname(os.path.abspath(__file__))

reg_extension = re.compile("\.py$")
files = [f for f in listdir(path) if reg_extension.search(f) != None and f != "__init__.py"]

files_dict = dict()

for filename in files:
    with open(path+"/"+filename, "r") as file:
        for line in file:
            if "category:" in line and "#" in line:
                cat = line.split(":")[1].replace("\n","").replace(" ","")
                update_dict(files_dict, cat, filename)
            break

options1 = [{"name": x, "value": i} for (i, x) in enumerate(files_dict.keys())]
options2 = [[{"name": y, "value": j} for (j, y) in enumerate(files_dict.get(z))] for z in files_dict.keys()]
name_is_ok = False

try:
    answer = interactive_cmd(options1, *options2)
    file_dir_path = os.path.abspath(sys.argv[0])
    selected_file = options2[answer[0]][answer[1]].get("name")
    template_path = os.path.join(path, selected_file)

    files_in_target_dir = [f for f in os.listdir(file_dir_path) if not f.startswith("~$")]

    while selected_file in files_in_target_dir:
        print("WARNING:\tScript called" + wrap_from_both_sides(selected_file, " \"") + "already exists in current directory.")
        overwrite = input("\nDo you want to overwrite it? [Y/n] \n")

        if overwrite == "Y":
            break

        elif overwrite == "n":
            rename = input("\nDo you want to save it with different name? [Y/n] \n")

            if rename == "Y":
                while name_is_ok is False:
                    new_file_name = input("\nEnter new script name: \n") or selected_file.replace(".py", "_new.py")

                    if ".py" in new_file_name:
                        name_is_ok = True
                        selected_file = new_file_name
                    else:
                        print("ERROR:\tInvalid file name. It doesn't seem to have the \".py\" extension. Reinitializing ...\n")
                        continue

            else:
                print("\nINFO:\tScript archiving canceled.")
                sys.exit(0)

        else:
            print("INFO:\tArgument hasn't been recognised. Script archiving canceled.")
            sys.exit(0)

    target_path = os.path.join(file_dir_path, selected_file)
    print("INFO:\tFile" + wrap_from_both_sides(selected_file, " \"") + "will be copied" +
          "\n\t> from"+wrap_from_both_sides(template_path, " \"") +
          "\n\t> to" + wrap_from_both_sides(target_path, " \"")
          )

    confirmation = input("\nPress Enter to proceed. \n")

    if confirmation == "":
        copy(template_path, target_path)
        print("INFO:\tScript successfully copied to" + wrap_from_both_sides(target_path, " \""))
        sys.exit(0)

    else:
        print("INFO:\tScript archiving canceled.")
        sys.exit(0)

except IOError:
    pass

except SystemExit:
    pass

except:
    static_cmd(options1, *options2)






from contextlib import contextmanager
from threading import local
import numpy as np

# Global context for energy units
_energy_context = local()
_energy_context.current_units = "kcal/mol"

@contextmanager
def energy_units(units):
    """Context manager for setting energy units temporarily."""
    previous_units = getattr(_energy_context, 'current_units', "kcal/mol")
    _energy_context.current_units = units
    try:
        yield
    finally:
        _energy_context.current_units = previous_units

def get_current_energy_units():
    """Get the current energy units from context."""
    return getattr(_energy_context, 'current_units', "kcal/mol")

@np.vectorize
def convert_nm_to_rcm(x):
    if x == 0:
        return 0
    else:
        return 1e7 / x




import numpy as np
import matplotlib.pyplot as plt
import colour
from colour_of_molecule.classes.classes import EnergyRange, Energy, EnergyAxis
from colour_of_molecule.utils.energy_units import get_current_energy_units
from colour_of_molecule.plotting.lineshapes import gaussian_profile

def abslines_to_molar_abs(file, title="Plot1", show_plot=False):
    abslines = file.abs_lines
    plot_range = file._plot_range
    normalize = file.normalize_absorption_spectrum
    bands = list(map(lambda k: k.energy, abslines))
    fs = list(map(lambda l: l.oscillator_strength, abslines))

    # Basic check that we have the same number of bands and oscillator strengths
    if len(bands) != len(fs):
        raise Exception('ERROR:   Number of bands does not match the number of oscillator strengths.')

    x = plot_range.energy_axis
    converted_x = x.in_current_units()
    composite = np.zeros_like(x.value)

    for i, band in enumerate(bands):
        # lineshape always calculated in linear energy units (kcal/mol - default unit)
        peak = gaussian_profile(x.value,
                                float(band.value), 
                                float(fs[i]), 
                                file._fwhm.value)
        composite += peak

    if show_plot == True:
        figg, axx = plt.subplots()
        axx.plot(x.value, composite)
        plt.xlabel('$E$ / {}'.format(get_current_energy_units()))
        plt.ylabel('$\epsilon$ / L mol$^{-1}$ cm$^{-1}$')
        plt.show()

    if normalize is True:
        from colour_of_molecule.analysis.common_tools import normalize_list
        composite = np.array(normalize_list(composite))

    x_sorted = np.sort(converted_x)
    y_sorted = composite[np.argsort(converted_x)]

    data = colour.SpectralDistribution(data=y_sorted, domain=x_sorted, name=title)

    return data


def molar_abs_to_transmittance(spectrum, OD=0.15):
    import colour
    import numpy as np

    val = spectrum.values
    wav = spectrum.wavelengths
    title = spectrum.name
    export = list()

    for j in range(0, len(val), 1):
        A = val[j]
        T = 10 ** (-A * OD)
        export.append(T)
    out = colour.SpectralDistribution(data=export, domain=wav, name=title)
    return out


def find_colour(spectrum):
    import colour

    col_map_f='CIE 1931 2 Degree Standard Observer'
    cmfs = colour.MSDS_CMFS[col_map_f]
    illuminant = colour.SDS_ILLUMINANTS['D65']

    energies = Energy(spectrum.wavelengths)
    wavelengths = np.array(energies.in_units("nm"), dtype=int)

    x_sorted = np.sort(wavelengths)
    y_sorted = spectrum.values[np.argsort(wavelengths)]

    wl_spectrum =colour.SpectralDistribution(data=y_sorted, domain=x_sorted, name=spectrum.name)
    XYZ = colour.sd_to_XYZ(wl_spectrum, cmfs, illuminant)

    rgb = colour.XYZ_to_sRGB(XYZ / 100)
    RGB = np.clip(rgb, 0, 1)

    return RGB


def find_colour_single(wl):
    import colour
    from colour.colorimetry import wavelength_to_XYZ

    if wl < 360 or wl > 830:
        RGB = (0, 0, 0)
    else:
        XYZ = wavelength_to_XYZ(wl)
        RGB = colour.XYZ_to_sRGB(XYZ)
        for i in range(0, 3, 1):
            if RGB[i] < 0:
                RGB[i] = 0
            if RGB[i] > 1:
                RGB[i] = 1
    return (RGB)





import re
from colour_of_molecule.classes import AbsLine, File, Energy
from colour_of_molecule.analysis.common_tools import homo_lumo
from colour_of_molecule.utils.energy_units import get_current_energy_units


class File_orca(File):
    class AbsLines_getter:
        def __get__(self, instance, owner):
            data = Orca_file_to_abslines(instance)
            return data

    abs_lines = AbsLines_getter()


def Orca_file_to_abslines(log_file):
    shift = log_file._shift

    with open(log_file.path, "r") as file:

        fs = list()
        wav = list()

        # parameters = list()
        state_no = 0
        states = list()
        state_numbers = list()
        orbitals = list()
        all_orbitals = list()
        num_of_el = list()

        reg_state = re.compile("^STATE\s*\d+\:")
        reg_line = re.compile("^-+")
        # reg_charge = re.compile("^\s?Charge.*\sMultiplicity")

        reg_MOs = re.compile("^N(\(Alpha\)|\(Beta\))\s*\:")
        reg_num = re.compile("[\d.]+")

        edm = False
        edmf = False
        li = 0
        HOMO = int()

        for line in file:
            if reg_MOs.search(line) is not None:
                num_of_el.append(reg_num.findall(line)[0])
                if len(num_of_el) == 2:
                    HOMO = int(round(float(max(num_of_el))))

            if state_no != 0:
                if "->" in line:
                    numerals = reg_num.findall(line)
                    orb_nos = numerals[:2]
                    amplitude = float(numerals[3])
                    ltt = list(map(lambda z: homo_lumo(HOMO, z), orb_nos))
                    ltt.append(amplitude)
                    orbitals.append(ltt)
                else:
                    all_orbitals.append(orbitals)
                    state_numbers.append(state_no)
                    orbitals = []
                    state_no = 0

            if reg_state.search(line) is not None and state_no == 0:
                state_no = int(reg_num.findall(line)[0])
                states.append(state_no)

            if edm is True:
                if li == 2:
                    edmf = True
                    edm = False
                elif reg_line.search(line) is not None:
                    li += 1

            if edmf is True:
                if reg_num.search(line) is not None:
                    nums = reg_num.findall(line)
                    energy = Energy(float(nums[2]), "nm") + shift
                    wavelength = energy.in_units("nm")
                    wav.append(wavelength)
                    fs.append(float(nums[3]))
                else:
                    edm = False
                    edmf = False

            if "ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS" in line:
                edm = True

    if wav is []:
        raise Exception("Error in file import. Check the encoding of .txt file and eventually change it to ANSI.")

    if shift.value != 0:
        print("Shift applied to wavelengths:\n  {} {}".format(shift.in_current_units(), get_current_energy_units()))

    output = list(map(lambda i, j, k: AbsLine(i, j, k), wav, fs, all_orbitals))

    return output




from ..classes.classes import Energy, FontSettings
import numpy as np
from matplotlib import rcParams
from matplotlib import pyplot as plt
from colour_of_molecule.classes.classes import File, Energy
from colour_of_molecule.utils.energy_units import get_current_energy_units
from colour_of_molecule.analysis.spectrum import find_colour_single

def plot_single_spectrum(file, save="", size=(6,3), dpi=400, rainbow=True,
                         title="", fonts=FontSettings(),
                         yaxis_label="relative absorbance",
                         lines_show=True, lines_ratio=(14, 1), lines_colours=True, 
                         lines_width=1.2, lines_lim=0.0001):
    #sanity check:
    if not isinstance(file, File):
        raise Exception("ERROR:\tUnrecognized input. First argument has to be derived from class \"File\".")

    if title != "":
        file.plot_title = title

    # from colour-science package:
    energies = Energy(file.molar_abs_spectrum.wavelengths, get_current_energy_units())
    values = file.molar_abs_spectrum.values

    xaxis = energies.in_current_units()
    yaxis = values
    
    plot_range = file._plot_range

    rcParams['font.size'] = fonts.sizedict['all']
    rcParams['font.family'] = fonts.fontdict['all']

    nrows = 1 if lines_show is False else 2
    gskw = dict(dict(height_ratios=list(lines_ratio), hspace=0.0)) if lines_show is True else None
    fig, ax = plt.subplots(nrows=nrows, figsize=size, dpi=dpi, facecolor='w', edgecolor='k',
                           gridspec_kw=gskw, sharex=True)

    boo = type(ax) == np.ndarray
    ax0 = ax[0] if boo is True else ax
    ax1 = ax[1] if boo is True else ax

    xaxis_label = "{} [{}]".format("energy", get_current_energy_units())

    ax1.set_xlabel(xaxis_label)
    ax1.set_xlim(*sorted(plot_range.in_current_units()))
    ax0.set_ylabel(yaxis_label)
    ax0.set_ylim(0, max(yaxis))
    ax0.ticklabel_format(axis="y", style="sci", scilimits=(-1, 1))
    ax0.locator_params(axis='y', nbins=5)

    ax0.plot(xaxis, yaxis, linewidth=1.2, color='k')

    if lines_show is True:
        abslines = file.abs_lines
        for absline in abslines:
            wv = absline.energy.in_current_units()
            if lines_colours is False:
                col = 'k'
            else:
                if absline.oscillator_strength > lines_lim:
                    col = find_colour_single(wv)
                else:
                    col = 'k'
            ax1.plot([wv, wv],[0, 1], color=col, linewidth=lines_width)
        plt.setp(ax1.get_yticklabels(), visible=False)
        plt.setp(ax1.get_yticklines(), visible=False)

    if rainbow is True:
        add_rainbow(ax0, file)

    for label in [*ax1.get_xticklabels(), *ax0.get_yticklabels()]:
        label.set_fontproperties(fonts.fonts['axis_tick_labels'])

    plot_title = file.plot_title if file.plot_title != "" else "File: {0}".format(file.filename)
    ax0.set_title(plot_title)

    plt.tight_layout()

    if save != "":
        if isinstance(save, str):
            from colour_of_molecule.analysis.common_tools import file_saver
            fpath = file_saver(save)
            if fpath != save:
                print("INFO:\tThe provided filename already exists. The file will be saved in \"{new}\" instead.".format(new=fpath))
            fig.savefig(fpath, dpi=dpi)
            print("INFO:\tImage saved successfully in \"{path}\".".format(path=fpath))

    plt.show()


def add_rainbow(axis, file, opacity=100):
    energies = Energy(file.molar_abs_spectrum.wavelengths, get_current_energy_units())
    wavelengths = energies.in_units("nm")
    values = file.molar_abs_spectrum.values

    xaxis = energies.in_current_units()
    yaxis = values

    # sanity check:
    if not hasattr(axis, 'plot') and not hasattr(axis, 'add_patch'):
        raise Exception("ERROR:\tFirst argument needs to have method \"plot\".")

    from colour.plotting import XYZ_to_plotting_colourspace, filter_cmfs, CONSTANTS_COLOUR_STYLE
    from colour.colorimetry import CCS_ILLUMINANTS, wavelength_to_XYZ
    from colour.utilities import first_item
    from colour.algebra import normalise_maximum
    from matplotlib.patches import Polygon

    col_map_f = "CIE 1931 2 Degree Standard Observer"

    cmfs = first_item(filter_cmfs(col_map_f).values())
    cmfs_sele = (wavelengths > cmfs.shape.start) & (wavelengths < cmfs.shape.end)
    wl_cmfs = wavelengths[cmfs_sele]
    eng_cmfs = xaxis[cmfs_sele]

    xyz = wavelength_to_XYZ(wl_cmfs, cmfs)
    clr = XYZ_to_plotting_colourspace(
        xyz,
        CCS_ILLUMINANTS[col_map_f]['E'],
        apply_cctf_encoding=False)

    clr = normalise_maximum(clr)
    clr = CONSTANTS_COLOUR_STYLE.colour.colourspace.cctf_encoding(clr)

    polygon = Polygon(
        np.vstack([
            [xaxis[0], 0],
            np.array([xaxis, yaxis]).T.tolist(),
            [xaxis[-1], 0],
        ]),
        facecolor='none',
        edgecolor='none')
    axis.add_patch(polygon)

    if opacity < 100:
        padding = 0
    else:
        padding = 0.1

    widths = np.abs(xaxis[1:] - xaxis[:-1])
    widths =np.append(widths, widths[-1])
    widths_cmfs = widths[cmfs_sele]

    for dom, col, w in [(xaxis - padding, 'black', widths), (eng_cmfs, clr, widths_cmfs)]:
        axis.bar(
            x=dom,
            height=max(values),
            width=w + padding,
            color=col,
            align='edge',
            alpha=opacity/100,
            clip_path=polygon
        )

    pass


def create_label(list1, list2):
    if type(list1) != list:
        list1 = [list1]
    if type(list2) != list:
        list2 = [list2]

    def equal_length(list1, list2):
        len1 = len(list1)
        len2 = len(list2)
        dif = len1 - len2
        if dif < 0:
            list1.extend([""] * abs(dif))
        elif dif > 0:
            list2.extend([""] * abs(dif))
        return list1, list2

    list1, list2 = equal_length(list1, list2)

    len1 = max([len(i) for i in list1])
    len2 = max([len(i) for i in list2])

    part = '{0: {align}{length}}'

    lines = str()
    for i, (m, n) in enumerate(zip(list1, list2)):
        if i == 0:
            line = part.format(" $\\rightarrow$ ", align='^', length=4).join(
                (part.format(m, align='>', length=len1), part.format(n, align='<', length=len2)))
            lines = line
        else:
            line = part.format("", align='^', length=4).join(
                (part.format(m, align='>', length=len1), part.format(n, align='<', length=len2)))
            lines = lines + "\n" + line

    return lines


def plot_abs_lines(file, save="", size=(7,7), dpi=200, fonts=FontSettings(),
                   title="",
                   xaxis_label="wavelength [nm]", yaxis_label="$\epsilon$", yaxis_label_right="oscillator strength"):
    from colour_of_molecule.classes.classes import File
    from matplotlib import pyplot as plt
    from matplotlib import rcParams

    rcParams['font.size'] = fonts.sizedict['all']
    rcParams['font.family'] = fonts.fontdict['all']

    rcParams['xtick.bottom'] = rcParams['xtick.labelbottom'] = True
    rcParams['xtick.top'] = rcParams['xtick.labeltop'] = False

    # sanity check:
    if not isinstance(file, File):
        raise Exception("ERROR:\tFunction argument has to be of \"File\" class.")

    if title != "":
        file.plot_title = title

    ab_spectrum = file.molar_abs_spectrum
    wavelengths = ab_spectrum.wavelengths
    values = ab_spectrum.values
    wav_range = file.wavelength_range
    abs_lines = file.abs_lines

    fig2, axis1 = plt.subplots(dpi=dpi, figsize=size, facecolor='w', edgecolor='k')
    axis1.set_xlim(wav_range)
    axis2 = axis1.twinx()
    axis2.set_xlim(wav_range)

    axis1.locator_params(axis='y', nbins=5)
    axis1.ticklabel_format(axis="y", style="sci", scilimits=(-1, 1))

    axis1.plot(wavelengths, values, color='black', alpha=0.3, linewidth=0.6)

    def check_transition_amplitudes(absl):
        ls = [ab[2] for ab in absl if len(ab) >= 3]
        maximal = max(ls) if len(ls) > 0 else None
        return maximal

    for ab in abs_lines:
        text_label = ""
        maximal = check_transition_amplitudes(ab.transitions[0])
        #print("   ABL: ", ab.transitions[0])
        for args in ab.transitions[0]:

            length = len(args)
            if length >= 3:
                amplitude = args[2]
                c = " ({:.2f})".format(amplitude)
            else:
                amplitude = None
                c = ""
            a, b = args[:2]
            lab = create_label(a, b)
            #print("Lab: ", lab, c, amplitude, maximal, amplitude >= file.transition_minimal_amplitude or amplitude == maximal)
            if amplitude is not None:
                text_label += lab + c + "\n" if amplitude >= file.transition_minimal_amplitude or amplitude == maximal else ""
            else:
                text_label += lab + "\n"
        text_label = text_label[:-1]

        if ab.wavelength > wav_range[0] and ab.wavelength < wav_range[1]:
            axis2.plot([ab.wavelength, ab.wavelength], [0, ab.oscillator_strength], label=text_label, linewidth=2)

    axis1.set_ylabel(yaxis_label)
    axis1.set_xlabel(xaxis_label)
    axis2.set_ylabel(yaxis_label_right)

    for ax in [axis1, axis2]:
        for label in [*ax.get_xticklabels(), *ax.get_yticklabels()]:
            label.set_fontproperties(fonts.fonts['axis_tick_labels'])

    plot_title = file.plot_title if file.plot_title == "" else "File: {0}".format(file.filename)
    legend_title = file.legend_title if file.legend_title != "" \
        else "Electron transitions with amplitude greater than {:.2f} contributing to absorption lines:".format(file.transition_minimal_amplitude)

    plt.title(plot_title)
    axis2.legend(bbox_to_anchor=(0.5, -0.18), loc='upper center', ncol=3,
                 title=legend_title,
                 frameon=False, columnspacing=2, labelspacing=1.5, prop=fonts.fonts['legend'])

    plt.tight_layout()

    if save != "":
        if isinstance(save, str):
            from colour_of_molecule.analysis.common_tools import file_saver
            fpath = file_saver(save)
            if fpath != save:
                print("INFO:\tThe provided filename already exists. The file will be saved in \"{new}\" instead.".format(new=fpath))
            fig2.savefig(fpath, dpi=dpi)
            print("INFO:\tImage saved successfully in \"{path}\".".format(path=fpath))
    plt.show()


def get_colour(file, save="", size=(3,3), dpi=200, fonts=FontSettings(), title="RGB: {RGB}"):
    from colour_of_molecule.analysis.spectrum import find_colour
    from matplotlib import rcParams
    from matplotlib import pyplot as plt
    from colour_of_molecule.classes.classes import File

    # sanity check:
    if not isinstance(file, File):
        raise Exception("ERROR:\tUnrecognized input. First argument has to be derived from class \"File\".")

    RGB = find_colour(file.transmittance)

    print("RGB:\t{0}".format(RGB))

    rcParams['font.size'] = fonts.sizedict['all']
    rcParams['font.family'] = fonts.fontdict['all']
    fig, axes = plt.subplots(num=None, figsize=size, dpi=dpi, facecolor='w', edgecolor='k')

    axes.set_facecolor(RGB)
    axes.set_xlabel(title.format(RGB=", ".join(["%.3f" % i for i in RGB])), fontsize=fonts.sizedict['title'])
    plt.setp((axes.get_yticklabels(), axes.get_yticklines(), axes.get_xticklabels(), axes.get_xticklines()), visible=False)

    if save != "":
        if isinstance(save, str):
            from colour_of_molecule.analysis.common_tools import file_saver
            fpath = file_saver(save)
            if fpath != save:
                print("INFO:\tThe provided filename already exists. The file will be saved in \"{new}\" instead.".format(new=fpath))
            fig.savefig(fpath, dpi=dpi)
            print("INFO:\tImage saved successfully in \"{path}\".".format(path=fpath))

    plt.show()













import numpy as np

def gaussian_profile(x, band, strength, fwhm):
    stdev = fwhm / (2 * np.sqrt(2 * np.log(2)))
    bandshape = 9.2369174e7 * strength / stdev * np.exp(-((x - band)**2 / (2 * stdev**2)))
    return bandshape




### category: My_first_category

import os
import importlib.util
import sys

MODULE_PATH = "C:/Users/Michal Ptáček/PycharmProjects/colour_of_molecule_v2/colour_of_molecule/__init__.py"
MODULE_NAME = "colour_of_molecule"

spec = importlib.util.spec_from_file_location(MODULE_NAME, MODULE_PATH)
module = importlib.util.module_from_spec(spec)
sys.modules[spec.name] = module 
spec.loader.exec_module(module)

#print(os.path.dirname(os.path.abspath(__file__)))

from colour_of_molecule import templates


